# %%
import json
import os
import os.path as osp

import networkx as nx
import numpy as np
import pandas as pd
import torch
from gene_graph.sc_graph import get_gene_list
from sklearn.preprocessing import OneHotEncoder, OrdinalEncoder
from torch_geometric.data import Data
from torch_geometric.utils import from_networkx
from tqdm import tqdm


# TODO check if theses are actually DiGraphs.... when they were read in I thought they were multigraphs... check for all.
#
##############################--Encoders--####################################
class CategoricalEncoder(torch.nn.Module):
    def __init__(self, emb_dim, full_feature_dims):
        super(CategoricalEncoder, self).__init__()
        self.embedding_list = torch.nn.ModuleList()
        for i, dim in enumerate(full_feature_dims):
            emb = torch.nn.Embedding(dim, emb_dim)
            torch.nn.init.xavier_uniform_(emb.weight.data)
            self.embedding_list.append(emb)

    def forward(self, x):
        x_embedding = 0
        for i in range(x.shape[1]):
            x_embedding += self.embedding_list[i](x[:, i])

        return x_embedding


##############################--Encoders--####################################
##############################--Edge_dicts--####################################
def regulators_edge_dict() -> nx.DiGraph:
    print("Writing regulators features to graph")
    gene_list = get_gene_list()
    G = nx.Graph()
    G.add_nodes_from(gene_list)
    for gene in tqdm(gene_list):
        with open(f"data/preprocessed/node_ym_attrs/{gene}.json", "r") as f:
            data = json.load(f)
            regulators = [attrs["regulators"] for attrs in data.values()][0]
            # initialize lists
            regulators_identifier = []
            pubmed_id = []
            strain_background = []
            regulation_direction = []
            regulator_type = []
            annotation_type = []
            for reg in regulators:
                for k, v in reg.items():
                    if k == "regulatoryRegions.regulator.secondaryIdentifier":
                        regulators_identifier.append(v)
                    elif k == "regulatoryRegions.publications.pubMedId":
                        pubmed_id.append(v)
                    elif k == "regulatoryRegions.strainBackground":
                        strain_background.append(v)
                    elif k == "regulatoryRegions.annotationType":
                        annotation_type.append(v)
                    elif k == "regulatoryRegions.regulatorType":
                        regulator_type.append(v)
                    elif k == "regulatoryRegions.regulationDirection":
                        regulation_direction.append(v)

            for reg in zip(
                regulators_identifier,
                pubmed_id,
                strain_background,
                annotation_type,
                regulator_type,
                regulation_direction,
            ):
                if gene in G.nodes and reg[0] in G.nodes:
                    # Default order arranged according to tie_break_popular function
                    edge_dict = {
                        "pubmed_id": reg[1],
                        "strain_background": reg[2],
                        "annotation_type": reg[3],
                        "regulator_type": reg[4],
                        "regulation_direction": reg[5],
                    }
                    # removing None and replacing with string "None"
                    edge_dict = {k: v if v else "None" for k, v in edge_dict.items()}
                    # direction is regulator -> queried gene
                    G.add_edge(reg[0], gene, **edge_dict)
    return G


def protein_interactions_edge_dict() -> nx.DiGraph:
    print("Writing protein interactions features to graph")
    gene_list = get_gene_list()
    G = nx.Graph()
    G.add_nodes_from(gene_list)
    for gene in tqdm(gene_list):
        with open(f"data/preprocessed/node_ym_attrs/{gene}.json", "r") as f:
            data = json.load(f)
            physical_interactions = [
                attrs["physical_interactions"] for attrs in data.values()
            ][0]
            # initialize lists
            participant = []
            experiment_name = []
            annotation_type = []
            role = []
            detection_method_identifier = []
            for ppi in physical_interactions:
                for k, v in ppi.items():
                    if k == "interactions.participant2.secondaryIdentifier":
                        participant.append(v)
                    elif k == "interactions.details.experiment.name":
                        experiment_name.append(v)
                    elif k == "interactions.details.annotationType":
                        annotation_type.append(v)
                    elif (
                        k
                        == "interactions.details.experiment.interactionDetectionMethods.identifier"
                    ):
                        detection_method_identifier.append(v)
                    elif k == "interactions.details.role1":
                        role.append(v)
            for ppi in zip(
                participant,
                experiment_name,
                annotation_type,
                detection_method_identifier,
                role,
            ):
                if gene in G.nodes and ppi[0] in G.nodes:
                    # Default order arranged according to tie_break_popular function
                    edge_dict = {
                        "experiment_name": ppi[1],
                        "annotation_type": ppi[2],
                        "detection_method_identifier": ppi[3],
                        "role": ppi[4],
                    }
                    # Replace None with 'None'
                    edge_dict = {k: v if v else "None" for k, v in edge_dict.items()}
                    # Can Add ordering option.
                    # example
                    # desired_order_list = [5, 2, 4, 3, 1]
                    # reordered_dict = {k: sample_dict[k] for k in desired_order_list}
                    # direction is gene -> participant
                    G.add_edge(gene, ppi[0], **edge_dict)
    return G


def gene_interactions_edge_dict() -> nx.DiGraph:
    print("Writing gene interactions features to graph")
    gene_list = get_gene_list()
    G = nx.Graph()
    G.add_nodes_from(gene_list)
    for gene in tqdm(gene_list):
        with open(f"data/preprocessed/node_ym_attrs/{gene}.json", "r") as f:
            data = json.load(f)
            gene_interactions = [attrs["gene_interactions"] for attrs in data.values()][
                0
            ]
            participant = []
            experiment_name = []
            interaction_details_annotation_type = []
            interactions_details_phenotype = []
            interactions_details_role1 = []
            interactions_detection_methods_identifier = []
            p_value = []
            sgaScore = []
            for ggi in gene_interactions:
                for k, v in ggi.items():
                    if k == "interactions.participant2.secondaryIdentifier":
                        participant.append(v)
                    elif k == "interactions.details.experiment.name":
                        experiment_name.append(v)
                    elif k == "interactions.details.annotationType":
                        interaction_details_annotation_type.append(v)
                    elif (
                        k
                        == "interactions.details.experiment.interactionDetectionMethods.identifier"
                    ):
                        interactions_detection_methods_identifier.append(v)
                    elif k == "interactions.details.phenotype":
                        interactions_details_phenotype.append(v)
                    elif k == "interactions.details.role1":
                        interactions_details_role1.append(v)
                    elif k == "interactions.alleleinteractions.pvalue":
                        p_value.append(v)
                    elif k == "interactions.alleleinteractions.sgaScore":
                        sgaScore.append(v)

            for ggi in zip(
                participant,
                experiment_name,
                interaction_details_annotation_type,
                interactions_detection_methods_identifier,
                interactions_details_phenotype,
                interactions_details_role1,
                p_value,
                sgaScore,
            ):
                if gene in G.nodes and ggi[0] in G.nodes:
                    edge_dict = {
                        "experiment_name": ggi[1],
                        "interaction_details_annotation_type": ggi[2],
                        "interactions_detection_methods_identifier": ggi[3],
                        "interactions_details_phenotype": ggi[4],
                        "interactions_details_role1": ggi[5],
                        "p_value": ggi[6],
                        "sgaScore": ggi[7],
                    }
                    # handling None values
                    # replace None values in edge dict with 'None'
                    edge_dict = {k: v if v else "None" for k, v in edge_dict.items()}
                    # direction is gene -> participant... This should not matter for the gene graph because it should be symmetric
                    G.add_edge(gene, ggi[0], **edge_dict)
    return G


##############################--Edge_dicts--####################################
##############################--Miscellaneous--#################################
# This comes from n53... this entire file of functions still needs to be added
def agg_edge_data(G: nx.Graph) -> pd.DataFrame:
    # loop over G and print target, source, key, and data
    sources = []
    targets = []
    keys = []
    datas = []
    for source, target, key, data in G.edges(data=True, keys=True):
        sources.append(source)
        targets.append(target)
        keys.append(key)
        datas.append(data)

    df = pd.DataFrame(
        {"source": sources, "target": targets, "keys": keys, "data": datas}
    )
    df = (
        df.groupby(["source", "target"])["data"]
        .apply(list)
        .reset_index(name="edge_data")
    )
    df["edge_data_size"] = df["edge_data"].transform(len)
    return df


# TODO This belongs in other label data file
def get_synthetic_lethal() -> list:
    print("Getting synthetic lethality from yeastmined features")
    gene_list = get_gene_list()
    G = nx.Graph()
    G.add_nodes_from(gene_list)
    synthetic_lethal = []
    for gene in tqdm(gene_list):
        with open(f"data/preprocessed/node_ym_attrs/{gene}.json", "r") as f:
            data = json.load(f)
            gene_interactions = [attrs["gene_interactions"] for attrs in data.values()][
                0
            ]
            participant = []
            interactions_details_phenotype = []
            for ggi in gene_interactions:
                for k, v in ggi.items():
                    if k == "interactions.participant2.secondaryIdentifier":
                        participant.append(v)
                    elif k == "interactions.details.phenotype":
                        if v == "inviable":
                            interactions_details_phenotype.append(v)

            for ggi in zip(
                participant,
                interactions_details_phenotype,
            ):
                if gene in G.nodes and ggi[0] in G.nodes:
                    edge_dict = {
                        "interactions_details_phenotype": ggi[1],
                    }
                    genes_inviable = [
                        [[gene, ggi[0]], [v]] for k, v in edge_dict.items()
                    ]
                    synthetic_lethal.extend(genes_inviable)
    names = [i[0] for i in synthetic_lethal]
    df = pd.DataFrame(names)
    df.loc[df[0] > df[1], "names_sort"] = df[0] + "_" + df[1]
    df.loc[df[0] < df[1], "names_sort"] = df[1] + "_" + df[0]
    gene_0 = df.loc[df["names_sort"].drop_duplicates().index][0].to_list()
    gene_1 = df.loc[df["names_sort"].drop_duplicates().index][1].to_list()
    synthetic_lethal_clean = []
    for gene in zip(gene_0, gene_1):
        synthetic_lethal_clean.append([[gene[0], gene[1]], ["inviable"]])
    synthetic_lethal_clean
    return synthetic_lethal_clean


##############################--Miscellaneous--#################################
##############################--Node_attrs--####################################
def protein_half_life_df(gene_list: list) -> tuple:
    print("protein half life features:")
    df_protein_half_life = pd.DataFrame()
    for gene in tqdm(gene_list):
        with open(f"data/preprocessed/node_ym_attrs/{gene}.json", "r") as f:
            data = json.load(f)
        protein_half_life = [attrs["protein_half_life"] for attrs in data.values()][0]
        df = pd.DataFrame(protein_half_life)
        df_protein_half_life = pd.concat([df_protein_half_life, df])
    df_protein_half_life = df_protein_half_life[
        [
            "secondaryIdentifier",
            "proteins.proteinHalfLife.value",
            "proteins.proteinHalfLife.units",
            "proteins.proteinHalfLife.publication.pubMedId",
        ]
    ]
    df_clean = df_protein_half_life.copy()
    df_clean.loc[
        df_clean["proteins.proteinHalfLife.units"] == "min",
        "proteins.proteinHalfLife.value",
    ] = (
        df_protein_half_life[
            df_protein_half_life["proteins.proteinHalfLife.units"] == "min"
        ]["proteins.proteinHalfLife.value"]
        / 60
    )
    df_clean.loc[
        df_clean["proteins.proteinHalfLife.units"] == "min",
        "proteins.proteinHalfLife.units",
    ] = "hr"
    # can drop duplicates since rows are identical
    df_clean = df_clean.drop_duplicates()
    df_clean = df_clean[["secondaryIdentifier", "proteins.proteinHalfLife.value"]]
    df_fill_small = pd.DataFrame(
        df_clean[["proteins.proteinHalfLife.value"]].median()
    ).rename(columns={0: "fill_value"})
    fill_genes = list(set(gene_list) - set(df_clean["secondaryIdentifier"].to_list()))
    df_fill = pd.DataFrame(
        {
            "secondaryIdentifier": fill_genes,
            "proteins.proteinHalfLife.value": df_fill_small.loc[
                "proteins.proteinHalfLife.value"
            ].to_list()
            * len(fill_genes),
        }
    )
    return df_clean, df_fill


def median_protein_abundance_df(gene_list: list) -> tuple:
    print("median protein abundance features:")
    df_median_protein_abundance = pd.DataFrame()
    for gene in tqdm(gene_list):
        with open(f"data/preprocessed/node_ym_attrs/{gene}.json", "r") as f:
            data = json.load(f)
        median_protein_abundance = [
            attrs["median_protein_abundance"] for attrs in data.values()
        ][0]
        df = pd.DataFrame(median_protein_abundance)
        df_median_protein_abundance = pd.concat([df_median_protein_abundance, df])
    df_median_protein_abundance = df_median_protein_abundance.loc[
        (~df_median_protein_abundance["proteins.median"].isna())
    ]
    df_clean = df_median_protein_abundance.copy()
    df_clean.loc[
        df_clean["proteins.MAD"].isna(), "proteins.MAD"
    ] = df_median_protein_abundance["proteins.MAD"].mean()
    df_clean["proteins.MAD"].isna().sum()
    df_clean = df_clean[
        ["secondaryIdentifier", "qualifier", "proteins.median", "proteins.MAD"]
    ].drop_duplicates()
    qualifier = (
        pd.DataFrame(df_clean["qualifier"].value_counts().sort_values(ascending=False))
        .reset_index()
        .loc[0, "index"]
    )
    fill = df_clean[["proteins.median", "proteins.MAD"]].median()
    fill["qualifier"] = qualifier
    df_fill_small = pd.DataFrame({"fill_value": fill})
    fill_genes = list(set(gene_list) - set(df_clean["secondaryIdentifier"].to_list()))
    df_fill = pd.DataFrame(
        {
            "secondaryIdentifier": fill_genes,
            "proteins.median": df_fill_small.loc["proteins.median"].to_list()
            * len(fill_genes),
            "proteins.MAD": df_fill_small.loc["proteins.MAD"].to_list()
            * len(fill_genes),
            "qualifier": df_fill_small.loc["qualifier"].to_list() * len(fill_genes),
        }
    )
    return df_clean, df_fill


def chromosomal_location_df(gene_list: list) -> tuple:
    print("chromosomal location features:")
    df_chromosomal_location = pd.DataFrame()
    for gene in tqdm(gene_list):
        with open(f"data/preprocessed/node_ym_attrs/{gene}.json", "r") as f:
            data = json.load(f)
        chromosomal_location = [
            attrs["chromosomal_location"] for attrs in data.values()
        ][0]
        df = pd.DataFrame(chromosomal_location)
        df_chromosomal_location = pd.concat([df_chromosomal_location, df])
    df_chromosomal_location[
        df_chromosomal_location["secondaryIdentifier"].duplicated(keep=False)
    ].sort_values("secondaryIdentifier")
    df_clean = df_chromosomal_location.drop_duplicates()[
        [
            "secondaryIdentifier",
            "chromosome.primaryIdentifier",
            "chromosomeLocation.start",
            "chromosomeLocation.end",
            "chromosomeLocation.strand",
        ]
    ]
    df_gene_list = pd.DataFrame({"secondaryIdentifier": gene_list})
    df_clean = pd.merge(df_clean, df_gene_list, on="secondaryIdentifier", how="inner")
    return df_clean


def gene_node_dict(write_graph: bool = True) -> nx.DiGraph:
    file_name = "data/preprocessed/gene_reprs/yeastmine/node_reprs.gpickle"
    if osp.exists(file_name):
        print(f"reading graph from: {file_name}")
        G = nx.read_gpickle(file_name)
    else:
        print("Writing gene node features to graph")
        gene_list = get_gene_list()
        G = nx.Graph()
        G.add_nodes_from(gene_list)
        # features
        df_clean, df_fill = median_protein_abundance_df(gene_list)
        df_median_protein_abundance = pd.concat([df_clean, df_fill])
        df_clean, df_fill = protein_half_life_df(gene_list)
        df_protein_half_life = pd.concat([df_clean, df_fill])
        df_chromosomal_location = chromosomal_location_df(gene_list)
        # merge features
        df_temp = pd.merge(
            df_median_protein_abundance,
            df_protein_half_life,
            on="secondaryIdentifier",
            how="outer",
        )
        df_temp = pd.merge(
            df_temp, df_chromosomal_location, on="secondaryIdentifier", how="outer"
        )
        # write data to graph
        for row in df_temp.iterrows():
            gene = row[1]["secondaryIdentifier"]
            data = (row[1].drop("secondaryIdentifier")).to_dict()
            G.add_node(gene, **data)
        if write_graph:
            nx.write_gpickle(G, file_name)
    return G


###
# TODO add n53--graph_statistics.py
# - mainly need to look for writing "_digraph.gpickle"
###


def add_name_rename(G: nx.Graph):
    attrs = {}
    mapping = {}
    for i, name in enumerate(list(G.nodes())):
        attrs[name] = {"gene_name": name}
        mapping[name] = i
    nx.set_node_attributes(G, attrs)
    H = nx.relabel_nodes(G, mapping)
    return H


# TODO include ordinal encoder and lookup encoder
def node_reprs_to_data(
    enc: str = "lookup",
    emb_dim: int = 2,
    attr_included_names: list = None,
    save: bool = True,
):
    # One hot encoding protein interactions
    # default behavior is to include the all one_hot features inside conditional
    # names of features that could be potentially used
    file_name_torch = "data/preprocessed/gene_reprs/yeastmine/node_reprs.pt"
    # if osp.exists(file_name_torch):
    #     Gt = torch.load(file_name_torch)
    # else:
    attr_data_names = [
        "qualifier",
        "proteins.median",
        "proteins.MAD",
        "proteins.proteinHalfLife.value",
        "chromosome.primaryIdentifier",
        "chromosomeLocation.start",
        "chromosomeLocation.end",
        "chromosomeLocation.strand",
    ]

    if attr_included_names is None:
        attr_included_names = attr_data_names
    file_name_nx = "data/preprocessed/gene_reprs/yeastmine/node_reprs.gpickle"
    G = nx.read_gpickle(file_name_nx)
    G = add_name_rename(G)
    Gt = from_networkx(G)
    data = []
    for name in attr_included_names:
        if type(Gt[name]) != list:
            data.append(Gt[name].numpy())
        else:
            data.append(Gt[name])
    data = np.array(data).T
    df = pd.DataFrame(data)
    # "None" string used previous for plotting... this graph shouldn't have any nan values
    df.replace("None", np.nan, inplace=True)
    # replacements have already been done prior to this function.
    df.columns = attr_data_names
    df = df.astype(
        {
            "proteins.median": "float",
            "proteins.MAD": "float",
            "proteins.proteinHalfLife.value": "float",
            "chromosomeLocation.start": "float",
            "chromosomeLocation.end": "float",
        }
    )
    numeric_names = [
        "proteins.median",
        "proteins.MAD",
        "proteins.proteinHalfLife.value",
        "chromosomeLocation.start",
        "chromosomeLocation.end",
    ]
    # treat numeric features
    numeric_names_input = list(set(numeric_names).intersection(set(attr_data_names)))
    attr_numeric = df[numeric_names_input].to_numpy()
    attr_numeric = torch.tensor(attr_numeric)
    # treat one hot features
    categorical_names = [
        "qualifier",
        "chromosome.primaryIdentifier",
        "chromosomeLocation.strand",
    ]
    categorical_names_input = list(
        set(categorical_names).intersection(set(attr_data_names))
    )
    attr_one_hot = df[categorical_names_input].to_numpy()
    ###

    if (enc == "ordinal") or (enc == "lookup"):
        edge_attr_categorical = df[categorical_names_input].to_numpy()
        enc_ordinal = OrdinalEncoder()
        enc_ordinal.fit(edge_attr_categorical)
        edge_attr_categorical = enc_ordinal.transform(edge_attr_categorical)
        edge_attr_categorical_tensor = torch.tensor(edge_attr_categorical)
        if enc == "lookup":
            full_feature_dims = []
            for i in df[categorical_names_input]:
                # Get feature dims list
                # reference https://github.com/snap-stanford/ogb/blob/master/ogb/graphproppred/mol_encoder.py
                full_feature_dims.append(df[i].value_counts().shape[0])
            enc_lookup = CategoricalEncoder(
                emb_dim=emb_dim, full_feature_dims=full_feature_dims
            )
            edge_attr_categorical_tensor = enc_lookup(
                edge_attr_categorical_tensor.long()
            )
    elif enc == "onehot":
        # "None" string used previous for plotting
        df.replace("None", np.nan, inplace=True)
        edge_attr_categorical = (
            df[categorical_names_input].replace(np.nan, "None").to_numpy()
        )
        # Not sparse to so can be converted to torch tensor
        enc = OneHotEncoder(sparse=False)
        enc.fit(edge_attr_categorical)
        edge_attr_categorical_tensor = torch.tensor(
            enc.transform(edge_attr_categorical)
        )
    ###

    # combine one_hot and numeric
    attr_tensor = torch.concat((edge_attr_categorical_tensor, attr_numeric), dim=1)
    # concatenate tensors
    Gt.x = attr_tensor
    edge_attr_drop_names = attr_data_names
    # remove named features and keep one_hot features
    for name in edge_attr_drop_names:
        Gt[name] = None
    if save:
        torch.save(Gt, file_name_torch)
    return Gt


def unirep_to_data():
    gene_list = get_gene_list()
    base_path = "data/preprocessed/gene_reprs/unirep"
    file_names = os.listdir(base_path)
    assert len(file_names) == len(gene_list), "gene list and unirep genes do not match!"
    data = Data()
    dim = len(torch.load(osp.join(base_path, file_names[0])))
    x = torch.empty(len(file_names), dim)
    gene_name = []
    for i, gene in enumerate(gene_list):
        gene_name.append(gene)
        x[i] = torch.load(f"data/preprocessed/gene_reprs/unirep/{gene}.pt")
    data.x = x
    data.edge_index = torch.empty(2, 0)
    data.gene_name = gene_name
    data.num_nodes = len(gene_name)
    return data


def join_edge_node_reprs(data_edge: Data, data_node: Data):
    H = Data()
    if data_edge.gene_name == data_node.gene_name:
        print("joining data")
        H.x = data_node.x
        H.edge_index = data_edge.edge_index
        H.edge_attr = data_edge.edge_attr
        H.gene_name = data_edge.gene_name
        H.num_nodes = data_edge.num_nodes
    else:
        print("data did not have matching gene_name")
        H = None
    return H


##############################--Node_attrs--####################################
##############################--Edge_attrs--######################################


def nx_regulators_to_torch(
    enc: str = "lookup",
    emb_dim: int = 2,
    attr_included_names: list = None,
    save=True,
):
    # enc: ["lookup", "ordinal","one_hot"]
    # One hot encoding regulators
    # default behavior is to include the all one_hot features inside conditional
    edge_attr_data_names = [
        "strain_background",
        "annotation_type",
        "regulator_type",
        "regulation_direction",
    ]
    if attr_included_names is None:
        attr_included_names = edge_attr_data_names
    file_name = "data/preprocessed/graphs/regulators/raw/regulators_digraph.gpickle"
    G = nx.read_gpickle(file_name)
    G = add_name_rename(G)
    Gt = from_networkx(G)
    data = []
    for name in attr_included_names:
        data.append(Gt[name])
    data = np.array(data).T
    df = pd.DataFrame(data)
    if (enc == "ordinal") or (enc == "lookup"):
        edge_attr = df.to_numpy()
        enc_ordinal = OrdinalEncoder()
        enc_ordinal.fit(edge_attr)
        edge_attr = enc_ordinal.transform(edge_attr)
        edge_attr_tensor = torch.tensor(edge_attr)
        if enc == "lookup":
            full_feature_dims = []
            # Note on all ggi and nodes features need [categorical_names_input]
            for i in df:
                # Get feature dims list
                # reference https://github.com/snap-stanford/ogb/blob/master/ogb/graphproppred/mol_encoder.py
                full_feature_dims.append(df[i].value_counts().shape[0])
            enc_lookup = CategoricalEncoder(
                emb_dim=emb_dim, full_feature_dims=full_feature_dims
            )
            edge_attr_tensor = enc_lookup(edge_attr_tensor.long())
    elif enc == "onehot":
        # "None" string used previous for plotting
        df.replace("None", np.nan, inplace=True)
        edge_attr = df.to_numpy()
        # Not sparse to so can be converted to torch tensor
        enc = OneHotEncoder(sparse=False)
        enc.fit(edge_attr)
        edge_attr_tensor = torch.tensor(enc.transform(edge_attr))
    Gt.edge_attr = edge_attr_tensor
    edge_attr_drop_names = [
        "strain_background",
        "annotation_type",
        "regulator_type",
        "regulation_direction",
    ] + ["pubmed_id"]
    # remove named features and keep one_hot features
    for name in edge_attr_drop_names:
        Gt[name] = None
    if save:
        torch.save(Gt, file_name.split(".")[0] + ".pt")
    return Gt


def nx_protein_interactions_to_torch(
    enc: str = "lookup",
    emb_dim: int = 2,
    attr_included_names: list = None,
    save=True,
):
    # One hot encoding protein interactions
    # default behavior is to include the all one_hot features inside conditional
    # names of features that could be potentially used
    edge_attr_data_names = ["annotation_type", "detection_method_identifier", "role"]
    if attr_included_names is None:
        attr_included_names = edge_attr_data_names
    file_name = "data/preprocessed/graphs/protein_interactions/raw/protein_interactions_digraph.gpickle"
    G = nx.read_gpickle(file_name)
    G = add_name_rename(G)
    Gt = from_networkx(G)
    data = []
    for name in attr_included_names:
        data.append(Gt[name])
    data = np.array(data).T
    df = pd.DataFrame(data)
    ###
    if (enc == "ordinal") or (enc == "lookup"):
        edge_attr = df.to_numpy()
        enc_ordinal = OrdinalEncoder()
        enc_ordinal.fit(edge_attr)
        edge_attr = enc_ordinal.transform(edge_attr)
        edge_attr_tensor = torch.tensor(edge_attr)
        if enc == "lookup":
            full_feature_dims = []
            # Note on all ggi and nodes features need [categorical_names_input]
            for i in df:
                # Get feature dims list
                # reference https://github.com/snap-stanford/ogb/blob/master/ogb/graphproppred/mol_encoder.py
                full_feature_dims.append(df[i].value_counts().shape[0])
            enc_lookup = CategoricalEncoder(
                emb_dim=emb_dim, full_feature_dims=full_feature_dims
            )
            edge_attr_tensor = enc_lookup(edge_attr_tensor.long())
    elif enc == "onehot":
        # "None" string used previous for plotting
        df.replace("None", np.nan, inplace=True)
        edge_attr = df.to_numpy()
        # Not sparse to so can be converted to torch tensor
        enc = OneHotEncoder(sparse=False)
        enc.fit(edge_attr)
        edge_attr_tensor = torch.tensor(enc.transform(edge_attr))
    Gt.edge_attr = edge_attr_tensor
    ###
    edge_attr_drop_names = [
        "annotation_type",
        "detection_method_identifier",
        "role",
    ] + ["experiment_name"]
    # remove named features and keep one_hot features
    for name in edge_attr_drop_names:
        Gt[name] = None
    if save:
        torch.save(Gt, file_name.split(".")[0] + ".pt")
    return Gt


def nx_gene_interactions_to_torch(
    enc: str = "lookup",
    emb_dim: int = 2,
    attr_included_names: list = None,
    save=True,
):
    # One hot encoding protein interactions
    # default behavior is to include the all one_hot features inside conditional
    # names of features that could be potentially used
    edge_attr_data_names = [
        "interaction_details_annotation_type",
        # "interactions_detection_methods_identifier",
        # "interactions_details_phenotype",
        "interactions_details_role1",
        "p_value",
        "sgaScore",
    ]
    if attr_included_names is None:
        attr_included_names = edge_attr_data_names
    file_name = "data/preprocessed/graphs/gene_interactions/raw/gene_interactions_digraph.gpickle"
    G = nx.read_gpickle(file_name)
    G = add_name_rename(G)
    Gt = from_networkx(G)
    data = []
    for name in attr_included_names:
        data.append(Gt[name])
    data = np.array(data).T
    df = pd.DataFrame(data)
    # # "None" string used previous for plotting... this graph shouldn't have any nan values
    df.replace("None", np.nan, inplace=True)
    ## only different part compared to other graph functions.
    df.columns = edge_attr_data_names
    # replace with floats with median values
    df["sgaScore"] = df["sgaScore"].replace(np.nan, df["sgaScore"].median())
    df["p_value"] = df["p_value"].replace(np.nan, df["p_value"].median())
    # type case
    df = df.astype({"p_value": "float", "sgaScore": "float"})
    numeric_names = ["sgaScore", "p_value"]
    numeric_names_input = list(
        set(numeric_names).intersection(set(edge_attr_data_names))
    )
    edge_attr_numeric = df[numeric_names_input].to_numpy()
    edge_attr_numeric = torch.tensor(edge_attr_numeric)
    categorical_names = [
        "interaction_details_annotation_type",
        "interactions_detection_methods_identifier",
        "interactions_details_phenotype",
        "interactions_details_role1",
    ]
    categorical_names_input = list(
        set(categorical_names).intersection(set(edge_attr_data_names))
    )
    ###
    if (enc == "ordinal") or (enc == "lookup"):
        edge_attr_categorical = df[categorical_names_input].to_numpy()
        enc_ordinal = OrdinalEncoder()
        enc_ordinal.fit(edge_attr_categorical)
        edge_attr_categorical = enc_ordinal.transform(edge_attr_categorical)
        edge_attr_categorical_tensor = torch.tensor(edge_attr_categorical)
        if enc == "lookup":
            full_feature_dims = []
            for i in df[categorical_names_input]:
                # Get feature dims list
                # reference https://github.com/snap-stanford/ogb/blob/master/ogb/graphproppred/mol_encoder.py
                full_feature_dims.append(df[i].value_counts().shape[0])
            enc_lookup = CategoricalEncoder(
                emb_dim=emb_dim, full_feature_dims=full_feature_dims
            )
            edge_attr_categorical_tensor = enc_lookup(
                edge_attr_categorical_tensor.long()
            )
    elif enc == "onehot":
        # "None" string used previous for plotting
        df.replace("None", np.nan, inplace=True)
        edge_attr_categorical = (
            df[categorical_names_input].replace(np.nan, "None").to_numpy()
        )
        # Not sparse to so can be converted to torch tensor
        enc = OneHotEncoder(sparse=False)
        enc.fit(edge_attr_categorical)
        edge_attr_categorical_tensor = torch.tensor(
            enc.transform(edge_attr_categorical)
        )

    # # Not sparse to so can be converted to torch tensor
    # enc = OneHotEncoder(sparse=False)
    # enc.fit(edge_attr_categorical)
    # edge_attr_categorical = torch.tensor(enc.transform(edge_attr_categorical))
    ###
    # combine one_hot and numeric
    edge_attr_tensor = torch.concat(
        (edge_attr_categorical_tensor, edge_attr_numeric), dim=1
    )
    # concatenate tensors
    Gt.edge_attr = edge_attr_tensor
    edge_attr_drop_names = [
        "interaction_details_annotation_type",
        "interactions_detection_methods_identifier",
        "interactions_details_phenotype",
        "interactions_details_role1",
        "p_value",
        "sgaScore",
    ] + ["experiment_name"]
    # remove named features and keep one_hot features
    for name in edge_attr_drop_names:
        Gt[name] = None
    if save:
        torch.save(Gt, file_name.split(".")[0] + ".pt")
    return Gt


def nx_attr_to_torch(
    graph: str = None,
    enc: str = "lookup",
    emb_dim: int = 2,
    attr_included_names: list = None,
    save=True,
):
    print(f"Converting {graph} to torch Data:")
    if graph == "regulators":
        data = nx_regulators_to_torch(
            enc=enc,
            emb_dim=emb_dim,
            attr_included_names=attr_included_names,
            save=save,
        )
    elif graph == "protein_interactions":
        data = nx_protein_interactions_to_torch(
            enc=enc,
            emb_dim=emb_dim,
            attr_included_names=attr_included_names,
            save=save,
        )
    elif graph == "gene_interactions":
        data = nx_gene_interactions_to_torch(
            enc=enc,
            emb_dim=emb_dim,
            attr_included_names=attr_included_names,
            save=save,
        )
    elif graph == "yeastmine_node":
        data = node_reprs_to_data(
            enc=enc,
            emb_dim=emb_dim,
            attr_included_names=attr_included_names,
            save=save,
        )
    else:
        raise ValueError("graph not found")
    return data


def main():
    # synthetic_lethal = get_synthetic_lethal()
    # print(len(synthetic_lethal))
    # gene_node_dict()
    # # joining node features and edge features into one data object
    # data_reg = nx_attr_to_torch(graph="regulators", enc="lookup", emb_dim=2)
    # print(data_reg)
    # data_ppi = nx_attr_to_torch(graph="protein_interactions", enc="lookup", emb_dim=2)
    # print(data_ppi)
    # data_ggi = nx_attr_to_torch(graph="gene_interactions", enc="lookup", emb_dim=2)
    # print(data_ggi)
    data_yeastmine_node = nx_attr_to_torch(
        graph="yeastmine_node", enc="lookup", emb_dim=2
    )
    ############
    file_name_torch = "data/preprocessed/gene_reprs/yeastmine/node_reprs.pt"
    # load file_name_torch
    data = torch.load(file_name_torch)
    data.x
    #############3
    # print(data_yeastmine_node)
    # data_reg_full = join_edge_node_reprs(data_reg, data_yeastmine_node)
    # print(f"combined_node_attrs:{data_reg_full}")
    # # joining node features from yeastmine and unirep
    # data_unirep = unirep_to_data()
    # print(f"unirep_data:\n{data_unirep}")
    # if data_unirep.gene_name == data_yeastmine_node.gene_name:
    #     x_all = torch.concat((data_yeastmine_node.x, data_unirep.x), dim=1)
    #     print("yeastmine + unirep node features size:", x_all.shape)


if __name__ == "__main__":
    main()



# torchcell/multidigraph/uniprot_api_ec.py
# [[torchcell.multidigraph.uniprot_api_ec]]
# https://github.com/Mjvolk3/torchcell/tree/main/torchcell/multidigraph/uniprot_api_ec.py
# Test file: torchcell/multidigraph/test_uniprot_api_ec.py

import json
import os
import os.path as osp
import time
from asyncio import Task
from collections.abc import Callable
from typing import Any, Optional

import aiohttp
import requests
from attrs import define, field
from tqdm import tqdm

from torchcell.multidigraph.validation.locus_related.locus import (
    Alias,
    InteractionOverview,
    LocusData,
    LocusDataUrl,
    PhysicalExperiments,
    Qualities,
    Reference,
    validate_data,
)

# Copy paste from the s288c gff file
# Parent=YBL105C_id001,YBL105C_id002;Name=YBL105C_CDS;orf_classification=Verified;protein_id=UniProtKB:P24583


def get_uniprot_ec(uniprot_id):
    url = f"https://www.uniprot.org/uniprot/{uniprot_id}.txt"
    response = requests.get(url)
    for line in response.text.split("\n"):
        if "EC=" in line:
            return line.split("EC=")[1].split(";")[0].strip()


ec_number = get_uniprot_ec("P24583")
print("EC Number:", ec_number)

# Should output EC Number: 2.7.11.13
# This information is contained in the gbff file also.


def main():
    pass


if __name__ == "__main__":
    main()



from .dataset import Dataset
from torchcell.data import (
    ExperimentReferenceIndex,
    serialize_for_hashing,
    compute_sha256_hash,
)
from tqdm import tqdm


def compute_experiment_reference_index(
    dataset: Dataset,
) -> list[ExperimentReferenceIndex]:
    # Hashes for each reference
    print("Computing experiment_reference_index hashes...")
    reference_hashes = [
        compute_sha256_hash(serialize_for_hashing(data["reference"]))
        for data in tqdm(dataset)
    ]

    # Identify unique hashes
    unique_hashes = set(reference_hashes)

    # Initialize ExperimentReferenceIndex list
    reference_indices = []

    print("Finding unique references...")
    for unique_hash in tqdm(unique_hashes):
        # Create a boolean list where True indicates the presence of the unique reference
        index_list = [ref_hash == unique_hash for ref_hash in reference_hashes]

        # Find the corresponding reference object for the unique hash
        ref_index = reference_hashes.index(unique_hash)
        unique_ref = dataset[ref_index]["reference"]

        # Create ExperimentReferenceIndex object
        exp_ref_index = ExperimentReferenceIndex(reference=unique_ref, index=index_list)
        reference_indices.append(exp_ref_index)

    return reference_indices



###

## Pandas view of data
import pandas as pd

# Specify column names
column_names = [
    "seqid",
    "source",
    "type",
    "start",
    "end",
    "score",
    "strand",
    "phase",
    "attributes",
]

# Read the file
gff_data = pd.read_csv(
    "data/ncbi/s_cerevisiae/ncbi_dataset/data/GCF_000146045.2/genomic.gff",
    sep="\t",
    comment="#",
    names=column_names,
)

# Display the first few rows of the dataframe
print(gff_data.head())

# from Bio import SeqIO

# # specify the path to your .fna file
# path_to_fna = "data/ncbi/s_cerevisiae/ncbi_dataset/data/GCF_000146045.2/GCF_000146045.2_R64_genomic.fna"

# gene_list = []
# for record in SeqIO.parse(path_to_fna, "fasta"):
#     description = record.description.split(' ')
#     for i, desc in enumerate(description):
#         if desc == 'gene':
#             gene_list.append(description[i+1])

# # remove duplicates, if any
# gene_list = list(set(gene_list))

# print(gene_list)

#####3
import gffutils

# specify the path to your .fna file
path_to_gff = "data/ncbi/s_cerevisiae/ncbi_dataset/data/GCF_000146045.2/genomic.gff"

db = gffutils.create_db(
    path_to_gff,
    dbfn="data.db",
    force=True,
    keep_order=True,
    merge_strategy="merge",
    sort_attribute_values=True,
)

genes = []
for gene_feature in db.features_of_type("gene"):
    genes.append(
        gene_feature["Name"][0]
    )  # adjust this depending on the actual structure of your .gff file

####



# torchcell/dataset/experiment_dataset
# [[torchcell.dataset.experiment_dataset]]
# https://github.com/Mjvolk3/torchcell/tree/main/torchcell/dataset/experiment_dataset
# Test file: tests/torchcell/dataset/test_experiment_dataset.py


import json
import logging
import os
import os.path as osp
import pickle
import shutil
import zipfile
from collections.abc import Callable
import numpy as np
import lmdb
import pandas as pd
from torch_geometric.data import download_url
from tqdm import tqdm
from torch_geometric.data import Dataset
from torchcell.dataset import compute_experiment_reference_index
from torchcell.data import ExperimentReferenceIndex
from torchcell.datamodels import (
    BaseEnvironment,
    Genotype,
    FitnessExperiment,
    FitnessExperimentReference,
    FitnessPhenotype,
    Media,
    ReferenceGenome,
    SgaKanMxDeletionPerturbation,
    SgaNatMxDeletionPerturbation,
    SgaDampPerturbation,
    SgaSuppressorAllelePerturbation,
    SgaTsAllelePerturbation,
    Temperature,
    BaseExperiment,
    ExperimentReference,
)
from torchcell.sequence import GeneSet
from multiprocessing import Process, Queue
from typing import Iterable
from abc import ABC, abstractmethod
from functools import wraps

logging.basicConfig(level=logging.INFO)
log = logging.getLogger(__name__)


def post_process(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        result = func(self, *args, **kwargs)
        self.gene_set = self.compute_gene_set()
        self.experiment_reference_index
        return result

    return wrapper


class ExperimentDataset(Dataset, ABC):
    def __init__(
        self,
        root: str,
        skip_process_file_exist: bool = False,
        transform: Callable | None = None,
        pre_transform: Callable | None = None,
    ):
        self.preprocess_dir = osp.join(root, "preprocess")
        # TODO This is part of our custom Dataset to speed things up but should be removed when using pure pyg
        self.skip_process_file_exist = skip_process_file_exist
        self.env = None
        self._length = None
        self._gene_set = None
        self._df = None
        self._experiment_reference_index = None
        super().__init__(root, transform, pre_transform)

    @property
    @abstractmethod
    def experiment_class(self) -> BaseExperiment: ...

    @property
    @abstractmethod
    def reference_class(self) -> ExperimentReference: ...

    @property
    @abstractmethod
    def raw_file_names(self) -> list[str]: ...

    @property
    def processed_file_names(self) -> list[str]:
        return "lmdb"

    @post_process
    @abstractmethod
    def process(self):
        raise NotImplementedError

    @abstractmethod
    def download(self):
        raise NotImplementedError

    def _init_db(self):
        """Initialize the LMDB environment."""
        self.env = lmdb.open(
            osp.join(self.processed_dir, "lmdb"),
            readonly=True,
            lock=False,
            readahead=False,
            meminit=False,
        )

    def close_lmdb(self):
        if self.env is not None:
            self.env.close()
            self.env = None

    @property
    def df(self):
        if osp.exists(osp.join(self.preprocess_dir, "data.csv")):
            self._df = pd.read_csv(osp.join(self.preprocess_dir, "data.csv"))
        return self._df

    @abstractmethod
    def preprocess_raw(
        self, df: pd.DataFrame, preprocess: dict | None = None
    ) -> pd.DataFrame: ...

    @abstractmethod
    def create_experiment(self): ...

    def len(self) -> int:
        if self.env is None:
            self._init_db()

        with self.env.begin() as txn:
            length = txn.stat()["entries"]

        # Must be closed for dataloader num_workers > 0
        self.close_lmdb()

        return length

    def get(self, idx):
        if self.env is None:
            self._init_db()

        # Handling boolean index arrays or numpy arrays
        if isinstance(idx, (list, np.ndarray)):
            if isinstance(idx, list):
                idx = np.array(idx)
            if idx.dtype == np.bool_:
                idx = np.where(idx)[0]

            # If idx is a list/array of indices, return a list of data objects
            return [self.get_single_item(i) for i in idx]
        else:
            # Single item retrieval
            return self.get_single_item(idx)

    def get_single_item(self, idx):
        with self.env.begin() as txn:
            serialized_data = txn.get(f"{idx}".encode())
            if serialized_data is None:
                return None

            deserialized_data = pickle.loads(serialized_data)
            return deserialized_data

    @staticmethod
    def extract_systematic_gene_names(genotype):
        gene_names = []
        for perturbation in genotype.get("perturbations"):
            gene_name = perturbation.get("systematic_gene_name")
            gene_names.append(gene_name)
        return gene_names

    def compute_gene_set(self):
        gene_set = GeneSet()
        if self.env is None:
            self._init_db()

        with self.env.begin() as txn:
            cursor = txn.cursor()
            log.info("Computing gene set...")
            for key, value in tqdm(cursor):
                deserialized_data = pickle.loads(value)
                experiment = deserialized_data["experiment"]

                extracted_gene_names = self.extract_systematic_gene_names(
                    experiment["genotype"]
                )
                for gene_name in extracted_gene_names:
                    gene_set.add(gene_name)

        self.close_lmdb()
        return gene_set

    # Reading from JSON and setting it to self._gene_set
    @property
    def gene_set(self):
        if osp.exists(osp.join(self.preprocess_dir, "gene_set.json")):
            with open(osp.join(self.preprocess_dir, "gene_set.json")) as f:
                self._gene_set = GeneSet(json.load(f))
        # elif self._gene_set is None:
        #     raise ValueError(
        #         "gene_set not written during process. "
        #         "Please call compute_gene_set at the end of process."
        #     )
        else:
            self._gene_set = self.compute_gene_set()
        return self._gene_set

    @gene_set.setter
    def gene_set(self, value):
        if not value:
            raise ValueError("Cannot set an empty or None value for gene_set")
        with open(osp.join(self.preprocess_dir, "gene_set.json"), "w") as f:
            json.dump(list(sorted(value)), f, indent=0)
        self._gene_set = value

    @property
    def experiment_reference_index(self):
        index_file_path = osp.join(
            self.preprocess_dir, "experiment_reference_index.json"
        )

        if osp.exists(index_file_path):
            with open(index_file_path, "r") as file:
                data = json.load(file)
                # Assuming ReferenceIndex can be constructed from a list of dictionaries
                self._experiment_reference_index = [
                    ExperimentReferenceIndex(**item) for item in data
                ]
        elif self._experiment_reference_index is None:
            self._experiment_reference_index = compute_experiment_reference_index(self)
            with open(index_file_path, "w") as file:
                # Convert each ExperimentReferenceIndex object to dict and save the list of dicts
                json.dump(
                    [eri.model_dump() for eri in self._experiment_reference_index],
                    file,
                    indent=4,
                )

        self.close_lmdb()
        return self._experiment_reference_index

    @property
    @abstractmethod
    def experiment_class(self): ...

    @property
    @abstractmethod
    def reference_class(self): ...

    def transform_item(self, item):
        experiment_data = item["experiment"]
        reference_data = item["reference"]
        experiment = self.experiment_class(**experiment_data)
        reference = self.reference_class(**reference_data)
        return {"experiment": experiment, "reference": reference}

    def __repr__(self):
        return f"{self.__class__.__name__}({len(self)})"



# torchcell/datasets/scerevisiae/fitness_yeast9.py
# [[torchcell.datasets.scerevisiae.fitness_yeast9]]
# https://github.com/Mjvolk3/torchcell/tree/main/torchcell/datasets/scerevisiae/fitness_yeast9.py
# Test file: tests/torchcell/datasets/scerevisiae/test_fitness_yeast9.py




# torchcell/datasets/scerevisiae_/costanzo2016
# [[torchcell.datasets.scerevisiae.costanzo2016]]
# https://github.com/Mjvolk3/torchcell/tree/main/torchcell/datasets/scerevisiae/costanzo2016
# Test file: tests/torchcell/datasets/scerevisiae/test_costanzo2016.py

import logging
import os
import os.path as osp
import pickle
import shutil
import zipfile
from collections.abc import Callable
import lmdb
import pandas as pd
from torch_geometric.data import download_url
from tqdm import tqdm
from torchcell.datamodels import (
    BaseEnvironment,
    Genotype,
    FitnessExperiment,
    FitnessExperimentReference,
    FitnessPhenotype,
    Media,
    ReferenceGenome,
    SgaKanMxDeletionPerturbation,
    SgaNatMxDeletionPerturbation,
    SgaDampPerturbation,
    SgaSuppressorAllelePerturbation,
    SgaTsAllelePerturbation,
    Temperature,
    BaseExperiment,
    ExperimentReference,
)
from torchcell.dataset import ExperimentDataset, post_process


logging.basicConfig(level=logging.INFO)
log = logging.getLogger(__name__)


class SmfCostanzo2016Dataset(ExperimentDataset):
    url = (
        "https://thecellmap.org/costanzo2016/data_files/"
        "Raw%20genetic%20interaction%20datasets:%20Pair-wise%20interaction%20format.zip"
    )

    def __init__(
        self,
        root: str = "data/torchcell/smf_costanzo2016",
        transform: Callable | None = None,
        pre_transform: Callable | None = None,
        **kwargs,
    ):
        super().__init__(root, transform, pre_transform, **kwargs)

    @property
    def experiment_class(self) -> BaseExperiment:
        return FitnessExperiment

    @property
    def reference_class(self) -> ExperimentReference:
        return FitnessExperimentReference

    @property
    def raw_file_names(self) -> str:
        return "strain_ids_and_single_mutant_fitness.xlsx"

    def download(self):
        path = download_url(self.url, self.raw_dir)
        with zipfile.ZipFile(path, "r") as zip_ref:
            zip_ref.extractall(self.raw_dir)
        os.remove(path)

        # Move the contents of the subdirectory to the parent raw directory
        sub_dir = os.path.join(
            self.raw_dir,
            "Data File S1. Raw genetic interaction datasets: Pair-wise interaction format",
        )
        for filename in os.listdir(sub_dir):
            shutil.move(os.path.join(sub_dir, filename), self.raw_dir)
        os.rmdir(sub_dir)
        # remove any excess files not needed
        for file_name in os.listdir(self.raw_dir):
            # if the file name ends in .txt remove it
            if file_name.endswith(".txt"):
                os.remove(osp.join(self.raw_dir, file_name))

    @post_process
    def process(self):
        xlsx_path = osp.join(self.raw_dir, "strain_ids_and_single_mutant_fitness.xlsx")
        df = pd.read_excel(xlsx_path)
        df = self.preprocess_raw(df)
        (reference_phenotype_std_26, reference_phenotype_std_30) = (
            self.compute_reference_phenotype_std(df)
        )

        # Save preprocssed df - mainly for quick stats
        os.makedirs(self.preprocess_dir, exist_ok=True)
        df.to_csv(osp.join(self.preprocess_dir, "data.csv"), index=False)

        log.info("Processing SMF Files...")

        # Initialize LMDB environment
        env = lmdb.open(
            osp.join(self.processed_dir, "lmdb"),
            map_size=int(1e12),  # Adjust map_size as needed
        )

        with env.begin(write=True) as txn:
            for index, row in tqdm(df.iterrows(), total=df.shape[0]):
                experiment, reference = self.create_experiment(
                    row,
                    reference_phenotype_std_26=reference_phenotype_std_26,
                    reference_phenotype_std_30=reference_phenotype_std_30,
                )

                # Serialize the Pydantic objects
                serialized_data = pickle.dumps(
                    {
                        "experiment": experiment.model_dump(),
                        "reference": reference.model_dump(),
                    }
                )
                txn.put(f"{index}".encode(), serialized_data)

        env.close()

    def preprocess_raw(self, df: pd.DataFrame) -> pd.DataFrame:
        df["Strain_ID_suffix"] = df["Strain ID"].str.split("_", expand=True)[1]

        # Determine perturbation type based on Strain_ID_suffix
        df["perturbation_type"] = df["Strain_ID_suffix"].apply(
            lambda x: (
                "damp"
                if "damp" in x
                else (
                    "temperature_sensitive"
                    if "tsa" in x or "tsq" in x
                    else (
                        "KanMX_deletion"
                        if "dma" in x
                        else (
                            "NatMX_deletion"
                            if "sn" in x  # or "S" in x or "A_S" in x
                            else "suppression_allele" if "S" in x else "unknown"
                        )
                    )
                )
            )
        )

        # Create separate dataframes for the two temperatures
        df_26 = df[
            [
                "Strain ID",
                "Systematic gene name",
                "Allele/Gene name",
                "Single mutant fitness (26°)",
                "Single mutant fitness (26°) stddev",
                "perturbation_type",
            ]
        ].copy()
        df_26["Temperature"] = 26

        df_30 = df[
            [
                "Strain ID",
                "Systematic gene name",
                "Allele/Gene name",
                "Single mutant fitness (30°)",
                "Single mutant fitness (30°) stddev",
                "perturbation_type",
            ]
        ].copy()
        df_30["Temperature"] = 30

        # Rename the columns for fitness and stddev to be common for both dataframes
        df_26.rename(
            columns={
                "Single mutant fitness (26°)": "Single mutant fitness",
                "Single mutant fitness (26°) stddev": "Single mutant fitness stddev",
            },
            inplace=True,
        )

        df_30.rename(
            columns={
                "Single mutant fitness (30°)": "Single mutant fitness",
                "Single mutant fitness (30°) stddev": "Single mutant fitness stddev",
            },
            inplace=True,
        )

        # Concatenate the two dataframes
        combined_df = pd.concat([df_26, df_30], ignore_index=True)
        combined_df = combined_df.dropna()
        combined_df = combined_df.drop_duplicates()
        combined_df = combined_df.reset_index(drop=True)

        return combined_df

    @staticmethod
    def compute_reference_phenotype_std(df: pd.DataFrame):
        mean_stds = df.groupby("Temperature")["Single mutant fitness stddev"].mean()
        reference_phenotype_std_26 = mean_stds[26]
        reference_phenotype_std_30 = mean_stds[30]
        return reference_phenotype_std_26, reference_phenotype_std_30

    @staticmethod
    def create_experiment(row, reference_phenotype_std_26, reference_phenotype_std_30):
        # Common attributes for both temperatures
        reference_genome = ReferenceGenome(
            species="saccharomyces Cerevisiae", strain="s288c"
        )

        # Deal with different types of perturbations
        if "temperature_sensitive" in row["perturbation_type"]:
            genotype = Genotype(
                perturbations=[
                    SgaTsAllelePerturbation(
                        systematic_gene_name=row["Systematic gene name"],
                        perturbed_gene_name=row["Allele/Gene name"],
                        strain_id=row["Strain ID"],
                    )
                ]
            )
        elif "damp" in row["perturbation_type"]:
            genotype = Genotype(
                perturbations=[
                    SgaDampPerturbation(
                        systematic_gene_name=row["Systematic gene name"],
                        perturbed_gene_name=row["Allele/Gene name"],
                        strain_id=row["Strain ID"],
                    )
                ]
            )
        elif "KanMX_deletion" in row["perturbation_type"]:
            genotype = Genotype(
                perturbations=[
                    SgaKanMxDeletionPerturbation(
                        systematic_gene_name=row["Systematic gene name"],
                        perturbed_gene_name=row["Allele/Gene name"],
                        strain_id=row["Strain ID"],
                    )
                ]
            )
        elif "NatMX_deletion" in row["perturbation_type"]:
            genotype = Genotype(
                perturbations=[
                    SgaNatMxDeletionPerturbation(
                        systematic_gene_name=row["Systematic gene name"],
                        perturbed_gene_name=row["Allele/Gene name"],
                        strain_id=row["Strain ID"],
                    )
                ]
            )
        elif "suppression_allele" in row["perturbation_type"]:
            genotype = Genotype(
                perturbations=[
                    SgaSuppressorAllelePerturbation(
                        systematic_gene_name=row["Systematic gene name"],
                        perturbed_gene_name=row["Allele/Gene name"],
                        strain_id=row["Strain ID"],
                    )
                ]
            )

        environment = BaseEnvironment(
            media=Media(name="YEPD", state="solid"),
            temperature=Temperature(value=row["Temperature"]),
        )
        reference_environment = environment.model_copy()
        # Phenotype based on temperature
        smf_key = "Single mutant fitness"
        smf_std_key = "Single mutant fitness stddev"
        phenotype = FitnessPhenotype(
            graph_level="global",
            label="smf",
            label_error="smf_std",
            fitness=row[smf_key],
            fitness_std=row[smf_std_key],
        )

        if row["Temperature"] == 26:
            reference_phenotype_std = reference_phenotype_std_26
        elif row["Temperature"] == 30:
            reference_phenotype_std = reference_phenotype_std_30
        reference_phenotype = FitnessPhenotype(
            graph_level="global",
            label="smf",
            label_error="smf_std",
            fitness=1.0,
            fitness_std=reference_phenotype_std,
        )

        reference = FitnessExperimentReference(
            reference_genome=reference_genome,
            reference_environment=reference_environment,
            reference_phenotype=reference_phenotype,
        )

        experiment = FitnessExperiment(
            genotype=genotype, environment=environment, phenotype=phenotype
        )
        return experiment, reference


class DmfCostanzo2016Dataset(ExperimentDataset):
    url = (
        "https://thecellmap.org/costanzo2016/data_files/"
        "Raw%20genetic%20interaction%20datasets:%20Pair-wise%20interaction%20format.zip"
    )

    def __init__(
        self,
        root: str = "data/torchcell/smf_costanzo2016",
        subset_n: int = None,
        transform: Callable | None = None,
        pre_transform: Callable | None = None,
        **kwargs,
    ):
        self.subset_n = subset_n
        super().__init__(root, transform, pre_transform, **kwargs)

    def download(self):
        path = download_url(self.url, self.raw_dir)
        with zipfile.ZipFile(path, "r") as zip_ref:
            zip_ref.extractall(self.raw_dir)
        os.remove(path)

        # Move the contents of the subdirectory to the parent raw directory
        sub_dir = os.path.join(
            self.raw_dir,
            "Data File S1. Raw genetic interaction datasets: Pair-wise interaction format",
        )
        for filename in os.listdir(sub_dir):
            shutil.move(os.path.join(sub_dir, filename), self.raw_dir)
        os.rmdir(sub_dir)
        # remove any excess files not needed
        os.remove(osp.join(self.raw_dir, "strain_ids_and_single_mutant_fitness.xlsx"))

    @property
    def experiment_class(self) -> BaseExperiment:
        return FitnessExperiment

    @property
    def reference_class(self) -> ExperimentReference:
        return FitnessExperimentReference

    @property
    def raw_file_names(self) -> list[str]:
        return ["SGA_DAmP.txt", "SGA_ExE.txt", "SGA_ExN_NxE.txt", "SGA_NxN.txt"]

    def preprocess_raw(self, df: pd.DataFrame, preprocess: dict | None = None):
        log.info("Preprocess on raw data...")

        # Function to extract gene name
        def extract_systematic_name(x):
            return x.apply(lambda y: y.split("_")[0])

        # Extract gene names
        df["Query Systematic Name"] = extract_systematic_name(df["Query Strain ID"])
        df["Array Systematic Name"] = extract_systematic_name(df["Array Strain ID"])
        Temperature = df["Arraytype/Temp"].str.extract("(\d+)").astype(int)
        df["Temperature"] = Temperature
        df["query_perturbation_type"] = df["Query Strain ID"].apply(
            lambda x: (
                "damp"
                if "damp" in x
                else (
                    "temperature_sensitive"
                    if "tsa" in x or "tsq" in x
                    else (
                        "KanMX_deletion"
                        if "dma" in x
                        else (
                            "NatMX_deletion"
                            if "sn" in x  # or "S" in x or "A_S" in x
                            else "suppression_allele" if "S" in x else "unknown"
                        )
                    )
                )
            )
        )
        df["array_perturbation_type"] = df["Array Strain ID"].apply(
            lambda x: (
                "damp"
                if "damp" in x
                else (
                    "temperature_sensitive"
                    if "tsa" in x or "tsq" in x
                    else (
                        "KanMX_deletion"
                        if "dma" in x
                        else (
                            "NatMX_deletion"
                            if "sn" in x  # or "S" in x or "A_S" in x
                            else "suppression_allele" if "S" in x else "unknown"
                        )
                    )
                )
            )
        )
        means = df.groupby("Temperature")[
            "Double mutant fitness standard deviation"
        ].mean()

        # Extracting means for specific temperatures
        self.reference_phenotype_std_26 = means.get(26, None)
        self.reference_phenotype_std_30 = means.get(30, None)

        return df

    def process(self):
        os.makedirs(self.preprocess_dir, exist_ok=True)
        self._length = None
        # Initialize an empty DataFrame to hold all raw data
        df = pd.DataFrame()

        # Read and concatenate all raw files
        log.info("Reading and Concatenating Raw Files...")
        for file_name in tqdm(self.raw_file_names):
            file_path = os.path.join(self.raw_dir, file_name)

            # Reading data using Pandas; limit rows for demonstration
            df_temp = pd.read_csv(file_path, sep="\t")

            # Concatenating data frames
            df = pd.concat([df, df_temp], ignore_index=True)
        # Functions for data filtering... duplicates selection,
        df = self.preprocess_raw(df)

        # Subset
        if self.subset_n is not None:
            df = df.sample(n=self.subset_n, random_state=42).reset_index(drop=True)

        # Save preprocssed df - mainly for quick stats
        df.to_csv(osp.join(self.preprocess_dir, "data.csv"), index=False)

        log.info("Processing DMF Files...")

        # Initialize LMDB environment
        env = lmdb.open(
            osp.join(self.processed_dir, "lmdb"),
            map_size=int(1e12),  # Adjust map_size as needed
        )

        with env.begin(write=True) as txn:
            for index, row in tqdm(df.iterrows(), total=df.shape[0]):
                experiment, reference = self.create_experiment(
                    row,
                    reference_phenotype_std_26=self.reference_phenotype_std_26,
                    reference_phenotype_std_30=self.reference_phenotype_std_30,
                )

                # Serialize the Pydantic objects
                serialized_data = pickle.dumps(
                    {
                        "experiment": experiment.model_dump(),
                        "reference": reference.model_dump(),
                    }
                )
                txn.put(f"{index}".encode(), serialized_data)

        env.close()

    @staticmethod
    def create_experiment(row, reference_phenotype_std_26, reference_phenotype_std_30):
        # Common attributes for both temperatures
        reference_genome = ReferenceGenome(
            species="saccharomyces Cerevisiae", strain="s288c"
        )
        # genotype
        perturbations = []
        # Query
        if "temperature_sensitive" in row["query_perturbation_type"]:
            perturbations.append(
                SgaTsAllelePerturbation(
                    systematic_gene_name=row["Query Systematic Name"],
                    perturbed_gene_name=row["Query allele name"],
                    strain_id=row["Query Strain ID"],
                )
            )
        elif "damp" in row["query_perturbation_type"]:
            perturbations.append(
                SgaDampPerturbation(
                    systematic_gene_name=row["Query Systematic Name"],
                    perturbed_gene_name=row["Query allele name"],
                    strain_id=row["Query Strain ID"],
                )
            )
        elif "KanMX_deletion" in row["query_perturbation_type"]:
            perturbations.append(
                SgaKanMxDeletionPerturbation(
                    systematic_gene_name=row["Query Systematic Name"],
                    perturbed_gene_name=row["Query allele name"],
                    strain_id=row["Query Strain ID"],
                )
            )

        elif "NatMX_deletion" in row["query_perturbation_type"]:
            perturbations.append(
                SgaNatMxDeletionPerturbation(
                    systematic_gene_name=row["Query Systematic Name"],
                    perturbed_gene_name=row["Query allele name"],
                    strain_id=row["Query Strain ID"],
                )
            )
        elif "suppression_allele" in row["query_perturbation_type"]:
            perturbations.append(
                SgaSuppressorAllelePerturbation(
                    systematic_gene_name=row["Query Systematic Name"],
                    perturbed_gene_name=row["Query allele name"],
                    strain_id=row["Query Strain ID"],
                )
            )

        # Array
        if "temperature_sensitive" in row["array_perturbation_type"]:
            perturbations.append(
                SgaTsAllelePerturbation(
                    systematic_gene_name=row["Array Systematic Name"],
                    perturbed_gene_name=row["Array allele name"],
                    strain_id=row["Array Strain ID"],
                )
            )
        elif "damp" in row["array_perturbation_type"]:
            perturbations.append(
                SgaDampPerturbation(
                    systematic_gene_name=row["Array Systematic Name"],
                    perturbed_gene_name=row["Array allele name"],
                    strain_id=row["Array Strain ID"],
                )
            )
        elif "KanMX_deletion" in row["array_perturbation_type"]:
            perturbations.append(
                SgaKanMxDeletionPerturbation(
                    systematic_gene_name=row["Array Systematic Name"],
                    perturbed_gene_name=row["Array allele name"],
                    strain_id=row["Array Strain ID"],
                )
            )

        elif "NatMX_deletion" in row["array_perturbation_type"]:
            perturbations.append(
                SgaNatMxDeletionPerturbation(
                    systematic_gene_name=row["Array Systematic Name"],
                    perturbed_gene_name=row["Array allele name"],
                    strain_id=row["Array Strain ID"],
                )
            )

        elif "suppression_allele" in row["array_perturbation_type"]:
            perturbations.append(
                SgaSuppressorAllelePerturbation(
                    systematic_gene_name=row["Array Systematic Name"],
                    perturbed_gene_name=row["Array allele name"],
                    strain_id=row["Array Strain ID"],
                )
            )
        genotype = Genotype(perturbations=perturbations)
        # genotype
        environment = BaseEnvironment(
            media=Media(name="YEPD", state="solid"),
            temperature=Temperature(value=row["Temperature"]),
        )
        reference_environment = environment.model_copy()
        # Phenotype based on temperature
        dmf_key = "Double mutant fitness"
        dmf_std_key = "Double mutant fitness standard deviation"
        phenotype = FitnessPhenotype(
            graph_level="global",
            label="smf",
            label_error="smf_std",
            fitness=row[dmf_key],
            fitness_std=row[dmf_std_key],
        )

        if row["Temperature"] == 26:
            reference_phenotype_std = reference_phenotype_std_26
        elif row["Temperature"] == 30:
            reference_phenotype_std = reference_phenotype_std_30
        reference_phenotype = FitnessPhenotype(
            graph_level="global",
            label="smf",
            label_error="smf_std",
            fitness=1.0,
            fitness_std=reference_phenotype_std,
        )

        reference = FitnessExperimentReference(
            reference_genome=reference_genome,
            reference_environment=reference_environment,
            reference_phenotype=reference_phenotype,
        )

        experiment = FitnessExperiment(
            genotype=genotype, environment=environment, phenotype=phenotype
        )
        return experiment, reference


if __name__ == "__main__":
    from torchcell.loader import CpuExperimentLoader
    dataset = DmfCostanzo2016Dataset(
        root="data/torchcell/dmf_costanzo2016_subset_n_1000",
        subset_n=1000,
    )
    # dataset.experiment_reference_index
    # dataset[0]
    # serialized_data = dataset[0]["experiment"].model_dump()
    # new_instance = FitnessExperiment.model_validate(serialized_data)
    # print(new_instance == dataset[0]['experiment'])
    # Usage example
    print(len(dataset))
    print(dataset.experiment_reference_index)
    data_loader = CpuExperimentLoader(dataset, batch_size=1, num_workers=1)
    # Fetch and print the first 3 batches
    for i, batch in enumerate(data_loader):
        # batch_transformed = list(map(dataset.transform_item, batch))
        print(batch[0])
        print("---")
        if i == 3:
            break
    # Clean up worker processes
    data_loader.close()
    print("completed")

    ######
    # Single mutant fitness
    dataset = SmfCostanzo2016Dataset()
    print(len(dataset))
    # print(dataset[100])
    # serialized_data = dataset[100]["experiment"].model_dump()
    # new_instance = FitnessExperiment.model_validate(serialized_data)
    # print(new_instance == serialized_data)
    data_loader = CpuExperimentLoader(dataset, batch_size=1, num_workers=1)
    # Fetch and print the first 3 batches
    for i, batch in enumerate(data_loader):
        # batch_transformed = list(map(dataset.transform_item, batch))
        print(batch[0])
        print("---")
        if i == 3:
            break
    # Clean up worker processes
    data_loader.close()
    print("completed")



# torchcell/models/dcell.py
# [[torchcell.models.dcell]]
# https://github.com/Mjvolk3/torchcell/tree/main/torchcell/models/dcell.py
# Test file: torchcell/models/test_dcell.py

from collections import OrderedDict

import networkx as nx
import torch
import torch.nn as nn
from torch_geometric.data import Batch, Data
from torch_geometric.utils import from_networkx

from torchcell.graph import (
    SCerevisiaeGraph,
    filter_by_contained_genes,
    filter_by_date,
    filter_go_IGI,
    filter_redundant_terms,
)
from torchcell.sequence import GeneSet


class SubsystemModel(nn.Module):
    def __init__(self, input_size, output_size):
        super().__init__()
        self.output_size = output_size  # Store the output size as an attribute
        self.linear = nn.Linear(input_size, output_size)
        self.tanh = nn.Tanh()
        self.batchnorm = nn.BatchNorm1d(output_size)

    def forward(self, x):
        x = self.linear(x)
        x = self.tanh(x)
        x = self.batchnorm(x)
        return x


class DCell(nn.Module):
    def __init__(
        self,
        go_graph,
        subsystem_output_min: int = 20,
        subsystem_output_max_mult: float = 0.3,
    ):
        super().__init__()
        # HACK probably should reverse the edges in original graph
        go_graph = nx.reverse(go_graph, copy=True)
        self.subsystem_output_min = subsystem_output_min
        self.subsystem_output_max_mult = subsystem_output_max_mult
        self.go_graph = self.add_boolean_state(go_graph)
        self.subsystems = nn.ModuleDict()
        self.build_subsystems()

    @staticmethod
    def add_boolean_state(go_graph: nx.Graph) -> nx.Graph:
        for node_id in go_graph.nodes:
            if node_id == "GO:ROOT":
                go_graph.nodes[node_id]["mutant_state"] = torch.tensor(
                    [], dtype=torch.float32
                )
            else:
                subsystem_state_size = len(go_graph.nodes[node_id]["gene_set"])
                go_graph.nodes[node_id]["mutant_state"] = torch.ones(
                    subsystem_state_size, dtype=torch.float32
                )
        return go_graph

    def build_subsystems(self):
        nodes_sorted = list(nx.topological_sort(self.go_graph))

        current_nodes = []
        # Initialize subsystems based on descendants
        for node_id in nodes_sorted:
            descendants = set(nx.descendants(self.go_graph, node_id))
            # If there are no descendants, this is a true leaf node
            if not descendants:
                genes = self.go_graph.nodes[node_id]["gene_set"]
                self.subsystems[node_id] = SubsystemModel(
                    input_size=len(genes),
                    output_size=max(
                        self.subsystem_output_min,
                        int(self.subsystem_output_max_mult * len(genes)),
                    ),
                )
                current_nodes.append(node_id)

                # Process non-leaf nodes in reverse topological order
        # Reverse order to move up from leaf nodes towards root node
        for node_id in reversed(nodes_sorted):
            if node_id in self.subsystems:
                # Skip if subsystem is already initialized (as a leaf)
                continue

            # TODO pick up here with nodes parent to leaf nodes
            # Calculate the input size as the sum of the output sizes of the child subsystems
            children_output_sizes = sum(
                self.subsystems[child].output_size
                for child in self.go_graph.successors(node_id)
            )
            # Add the boolean state vector size for the current node
            # make empty gene set
            genes = self.go_graph.nodes[node_id].get("gene_set", [])
            total_input_size = children_output_sizes + len(genes)

            # Calculate the output size with a minimum of 20
            output_size = max(
                self.subsystem_output_min,
                int(self.subsystem_output_max_mult * len(genes)),
            )

            # Initialize the subsystem for the current node
            self.subsystems[node_id] = SubsystemModel(
                input_size=total_input_size, output_size=output_size
            )

    def calculate_input_size(self, node_id):
        # Sum output sizes of child subsystems and the boolean state vector
        input_size_from_children = sum(
            self.subsystems[child].output_size
            for child in self.go_graph.successors(node_id)
        )
        genes = self.go_graph.nodes[node_id]["gene_set"]
        return input_size_from_children + len(genes)

    def forward(self, batch: Batch):
        # HACK should probably move device to a more appropriate location
        subsystem_outputs = {}
        sorted_subsystems = reversed(list(nx.topological_sort(self.go_graph)))
        go_ids = [i for ids in batch.id for i in ids]
        device = batch.x.device
        for subsystem_name in sorted_subsystems:
            subsystem_model = self.subsystems[subsystem_name]

            # Find indices in the batch that correspond to this subsystem
            subsystem_indices = [
                i for i, go in enumerate(go_ids) if go == subsystem_name
            ]

            # Convert the mask to boolean if it's not already
            bool_mask = batch.mask.type(torch.bool)

            # Gather the mutant states for the subsystem
            mutant_states = torch.stack(
                [batch.x[idx][bool_mask[idx]].to(device) for idx in subsystem_indices]
            ).to(torch.float32)

            # Gather and concatenate outputs of child subsystems
            child_outputs = torch.tensor([], dtype=torch.float32).to(device)
            for child in self.go_graph.successors(subsystem_name):
                assert child in subsystem_outputs, "children must be processed first"
                child_output = subsystem_outputs[child].to(device)
                child_outputs = torch.cat((child_outputs, child_output), dim=1)

            # Concatenate child outputs with mutant states
            if len(child_outputs) > 0:
                subsystem_input = torch.cat((mutant_states, child_outputs), dim=1)
            else:
                subsystem_input = mutant_states

            # Compute the subsystem output
            subsystem_output = subsystem_model(subsystem_input.to(torch.float32))
            subsystem_outputs[subsystem_name] = subsystem_output

        return subsystem_outputs


class DCellLinear(nn.Module):
    def __init__(self, subsystems: nn.ModuleDict, output_size: int):
        super().__init__()
        self.output_size = output_size
        self.subsystem_linears = nn.ModuleDict()

        # Create a linear layer for each subsystem with the appropriate input size
        for subsystem_name, subsystem in subsystems.items():
            in_features = subsystem.output_size
            self.subsystem_linears[subsystem_name] = nn.Linear(
                in_features, self.output_size
            )

    def forward(self, subsystem_outputs: dict):
        # Initialize a dictionary to store the outputs for each subsystem
        linear_outputs = {}

        # Apply the linear transformation to each subsystem output
        for subsystem_name, subsystem_output in subsystem_outputs.items():
            transformed_output = self.subsystem_linears[subsystem_name](
                subsystem_output
            )
            linear_outputs[subsystem_name] = transformed_output

        return linear_outputs


def delete_genes(go_graph: nx.Graph, deletion_gene_set: GeneSet):
    G_mutant = go_graph.copy()
    for node in G_mutant.nodes:
        if node == "GO:ROOT":
            G_mutant.nodes[node]["mutant_state"] = torch.tensor([], dtype=torch.int32)
        else:
            gene_set = G_mutant.nodes[node]["gene_set"]
            # Replace the genes in the knockout set with 0
            G_mutant.nodes[node]["mutant_state"] = torch.tensor(
                [1 if gene not in deletion_gene_set else 0 for gene in gene_set],
                dtype=torch.int32,
            )
    return G_mutant


def dcell_from_networkx(G_mutant):
    G_mutant_copy = G_mutant.copy()

    # Initialize maximum length to zero
    max_length = 0
    # Find the maximum length of the 'mutant_state' tensor and simplify node data
    for node_id, node_data in G_mutant_copy.nodes(data=True):
        mutant_state = node_data.get(
            "mutant_state", torch.tensor([], dtype=torch.float32)
        )
        max_length = max(max_length, mutant_state.size(0))
        simplified_data = {"id": node_id, "mutant_state": mutant_state}
        G_mutant_copy.nodes[node_id].clear()
        G_mutant_copy.nodes[node_id].update(simplified_data)

    # Pad the mutant_state tensors to the maximum length and create masks
    mask_list = []
    for node_id, node_data in G_mutant_copy.nodes(data=True):
        mutant_state = node_data["mutant_state"]
        mask = torch.ones(max_length, dtype=torch.uint8)
        if mutant_state.size(0) < max_length:
            padding = max_length - mutant_state.size(0)
            padded_mutant_state = torch.cat([mutant_state, torch.full((padding,), -1)])
            node_data["mutant_state"] = padded_mutant_state
            # Update the mask to indicate which entries are actual data
            mask[-padding:] = 0
        mask_list.append(mask)

    # Convert the NetworkX graph to a PyTorch Geometric Data object
    data = from_networkx(G_mutant_copy)
    data.x = torch.stack(
        [node_data["mutant_state"] for _, node_data in G_mutant_copy.nodes(data=True)]
    )
    data.mask = torch.stack(mask_list)
    del data.mutant_state
    return data


def main():
    import os
    import os.path as osp
    import random

    import matplotlib.pyplot as plt
    import pandas as pd
    from dotenv import load_dotenv

    from torchcell.datasets.scerevisiae import (
        DmfCostanzo2016Dataset,
        SmfCostanzo2016Dataset,
    )
    from torchcell.losses import DCellLoss
    from torchcell.sequence.genome.scerevisiae.s288c import SCerevisiaeGenome

    load_dotenv()
    DATA_ROOT = os.getenv("DATA_ROOT")

    genome = SCerevisiaeGenome(
        data_root=osp.join(DATA_ROOT, "data/sgd/genome"), overwrite=True
    )
    graph = SCerevisiaeGraph(
        data_root=osp.join(DATA_ROOT, "data/sgd/genome"), genome=genome
    )
    dmf_dataset = DmfCostanzo2016Dataset(
        root=osp.join(DATA_ROOT, "data/scerevisiae/costanzo2016_1e5"),
        preprocess={"duplicate_resolution": "low_dmf_std"},
        # subset_n=100,
    )
    smf_dataset = SmfCostanzo2016Dataset(
        root=osp.join(DATA_ROOT, "data/scerevisiae/costanzo2016_smf"),
        preprocess={"duplicate_resolution": "low_std_both"},
        skip_process_file_exist_check=True,
    )
    gene_set = smf_dataset.gene_set.union(dmf_dataset.gene_set)
    #
    print(graph.G_go.number_of_nodes())
    G = graph.G_go.copy()

    # Filtering
    G = filter_by_date(G, "2017-07-19")
    print(f"After date filter: {G.number_of_nodes()}")
    G = filter_go_IGI(G)
    print(f"After IGI filter: {G.number_of_nodes()}")
    G = filter_redundant_terms(G)
    print(f"After redundant filter: {G.number_of_nodes()}")
    G = filter_by_contained_genes(G, n=2, gene_set=gene_set)
    print(f"After containment filter: {G.number_of_nodes()}")

    # Instantiate the model
    # dcell = DCell(go_graph=G)

    dcell = DCell(G)
    target = torch.rand(2, 1)
    # target = target.repeat_interleave(G.number_of_nodes())
    # Define the loss function
    criterion = DCellLoss()

    G_mutant = delete_genes(
        go_graph=dcell.go_graph, deletion_gene_set=GeneSet(("YDL029W", "YDR150W"))
    )
    # forward method
    G_mutant = dcell_from_networkx(G_mutant)

    batch = Batch.from_data_list([G_mutant, G_mutant])
    subsystem_outputs = dcell(batch)
    dcell_linear = DCellLinear(dcell.subsystems, output_size=1)
    output = dcell_linear(subsystem_outputs)
    # Compute the loss
    loss = criterion(output, target, dcell.parameters())

    # Backward pass
    loss.backward()

    print(f"Loss: {loss.item()}")
    print()


if __name__ == "__main__":
    main()



# torchcell/models/deep_set.py
# [[torchcell.models.deep_set]]
# https://github.com/Mjvolk3/torchcell/tree/main/torchcell/models/deep_set.py
# Test file: torchcell/models/test_deep_set.py

import torch
import torch.nn as nn
from torch_scatter import scatter_add

from torchcell.models.act import act_register


class DeepSet(nn.Module):
    def __init__(
        self,
        input_dim: int,
        node_layers: list[int],
        set_layers: list[int],
        dropout_prob: float = 0.2,
        norm: str = "batch",
        activation: str = "relu",
        skip_node: bool = False,  # Parameter to add skip connections in node_layers
        skip_set: bool = False,  # Parameter to add skip connections in set_layers
    ):
        super().__init__()

        assert norm in ["batch", "instance", "layer"], "Invalid norm type"
        assert activation in act_register.keys(), "Invalid activation type"

        self.skip_node = skip_node
        self.skip_set = skip_set

        def create_block(in_dim, out_dim, norm, activation):
            block = [nn.Linear(in_dim, out_dim)]
            if norm == "batch":
                block.append(nn.BatchNorm1d(out_dim))
            elif norm == "instance":
                block.append(nn.InstanceNorm1d(out_dim, affine=True))
            elif norm == "layer":
                block.append(nn.LayerNorm(out_dim))
            block.append(act_register[activation])
            return nn.Sequential(*block)

        in_dim = input_dim
        node_modules = []
        for out_dim in node_layers:
            node_modules.append(create_block(in_dim, out_dim, norm, activation))
            in_dim = out_dim
        self.node_layers = nn.ModuleList(node_modules)

        set_modules = []
        for out_dim in set_layers:
            set_modules.append(create_block(in_dim, out_dim, norm, activation))
            in_dim = out_dim
        set_modules.append(nn.Dropout(dropout_prob))
        self.set_layers = nn.ModuleList(set_modules)

    def node_layers_forward(self, x):
        """Process node features through node layers."""
        x_node = x
        for i, layer in enumerate(self.node_layers):
            out_node = layer(x_node)
            if self.skip_node and x_node.shape[-1] == out_node.shape[-1]:
                out_node = out_node + x_node  # Skip connection
            x_node = out_node
        return x_node

    def set_layers_forward(self, x_summed):
        """Process aggregated features through set layers."""
        x_set = x_summed
        for i, layer in enumerate(self.set_layers):
            out_set = layer(x_set)
            if self.skip_set and x_set.shape[-1] == out_set.shape[-1]:
                out_set = out_set + x_set  # Skip connection
            x_set = out_set
        return x_set

    def forward(self, x, batch):
        x_node = self.node_layers_forward(x)
        x_summed = scatter_add(x_node, batch, dim=0)
        x_set = self.set_layers_forward(x_summed)
        return x_node, x_set


def main():
    torch.autograd.set_detect_anomaly(True)

    # Model configuration
    input_dim = 10
    node_layers = [64, 32, 32, 32]
    set_layers = [16, 8, 8]

    model = DeepSet(
        input_dim,
        node_layers,
        set_layers,
        norm="layer",
        activation="tanh",
        skip_node=True,
        skip_set=True,
    )

    # Dummy data
    x = torch.rand(100, input_dim)
    batch = torch.cat([torch.full((20,), i, dtype=torch.long) for i in range(5)])

    # Forward pass
    x_nodes, x_set = model(x, batch)
    print(x_set.shape)
    print(x_nodes.shape)

    # Let's assume you want to predict some values for each set.
    # So, we'll create a dummy target tensor for demonstration purposes.
    target = torch.rand(5, set_layers[-1])

    # Simple mean squared error loss
    criterion = nn.MSELoss()
    loss = criterion(x_set, target)
    print("Loss:", loss.item())

    # Backpropagation
    model.zero_grad()
    loss.backward()
    print("Gradients computed successfully!")
    print(x_set.size())


if __name__ == "__main__":
    main()
